# Standard library
import abc

# This package
from .utils import batch_tasks

__all__ = ['BasePool']


def _callback_wrapper(callback, generator):
    for element in generator:
        callback(element)
        yield element


class BasePool(metaclass=abc.ABCMeta):
    """ A base class multiprocessing pool with a ``map`` method. """

    def __init__(self, **kwargs):
        self.rank = 0

    @staticmethod
    def enabled():
        return False

    def is_master(self):
        return self.rank == 0

    def is_worker(self):
        return self.rank != 0

    def wait(self):
        return

    @abc.abstractmethod
    def map(self, *args, **kwargs):
        return

    def batched_map(self, worker, tasks, *args, **kwargs):
        batches = batch_tasks(n_batches=self.size, arr=tasks)
        return self.map(worker, batches, *args, **kwargs)

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _call_callback(self, callback, generator):
        if callback is None:
            return generator
        return _callback_wrapper(callback, generator)



"""
I couldn't figure out how to get py.test and MPI to play nice together,
so this is a script that tests the MPIPool
"""

# Standard library
import random

import pytest

# Use full imports so we can run this with mpiexec externally
from schwimmbad.tests import TEST_MPI  # noqa
from schwimmbad.tests.test_pools import _function, _batch_function, isclose
from schwimmbad.mpi import MPIPool, MPI  # noqa


def _callback(x):
    pass


@pytest.mark.skip(True, reason="WTF")
def test_mpi():
    with MPIPool() as pool:
        all_tasks = [[random.random() for i in range(1000)]]

        # test map alone
        for tasks in all_tasks:
            results = pool.map(_function, tasks)
            for r1, r2 in zip(results, [_function(x) for x in tasks]):
                assert isclose(r1, r2)

            assert len(results) == len(tasks)

        # test map with callback
        for tasks in all_tasks:
            results = pool.map(_function, tasks, callback=_callback)
            for r1, r2 in zip(results, [_function(x) for x in tasks]):
                assert isclose(r1, r2)

            assert len(results) == len(tasks)

        # test batched map
        results = pool.batched_map(_batch_function, tasks)
        for r in results:
            assert all([isclose(x, 42.01) for x in r])


if __name__ == '__main__':
    test_mpi()



# Standard library
import time
import random

# Project
from ..serial import SerialPool
from ..multiprocessing import MultiPool
from ..jl import JoblibPool


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def _function(x):
    """
    Wastes a random amount of time, returns the number 42.01
    """
    time.sleep(random.random()*4E-4 + 1E-4)
    return 42.01


def _batch_function(batch):
    """
    Wastes a random amount of time, returns the number 42.01 - accepts a batch
    of tasks instead of a single task
    """
    results = []
    for x in batch:
        time.sleep(random.random()*4E-4 + 1E-4)
        results.append(42.01)
    return results


class PoolTestBase(object):

    all_tasks = [[random.random() for i in range(1000)]]

    def test_init_close(self):
        pool = self.PoolClass()
        pool.close()

    def _make_pool(self):
        return self.PoolClass()

    def test_map(self):
        pool = self._make_pool()

        for tasks in self.all_tasks:
            results = pool.map(_function, tasks)
            count = 0
            for r1,r2 in zip(results, [_function(x) for x in tasks]):
                assert isclose(r1, r2)
                count += 1

            assert count == len(tasks) # to make sure iterations through the loop happen!

        pool.close()

    def test_map_callback(self):
        pool = self._make_pool()

        for tasks in self.all_tasks:

            mylist = []
            def _callback(x):
                assert isclose(x, 42.01)
                mylist.append(x)

            results = pool.map(_function, tasks, callback=_callback)
            for r1,r2 in zip(results, [_function(x) for x in tasks]):
                assert isclose(r1, r2)
            assert len(mylist) == len(tasks)

        pool.close()

class TestSerialPool(PoolTestBase):
    def setup(self):
        self.PoolClass = SerialPool

class TestMultiPool(PoolTestBase):
    def setup(self):
        self.PoolClass = MultiPool

class TestJoblibPool(PoolTestBase):
    def setup(self):
        self.PoolClass = JoblibPool



"""
I couldn't figure out how to get py.test and MPI to play nice together,
so this is a script that tests the MPIPool
"""

# Standard library
import random
import sys

import pytest

# Use full imports so we can run this with mpiexec externally
from schwimmbad.tests import TEST_MPI  # noqa
from schwimmbad.tests.test_pools import _function, isclose
from schwimmbad.mpi import MPIPool, MPI  # noqa


def _callback(x):
    pass


@pytest.mark.skip(True, reason="WTF")
def test_mpi_with_dill():
    pool = MPIPool(use_dill=True)

    pool.wait(lambda: sys.exit(0))

    all_tasks = [[random.random() for i in range(1000)]]

    # test map alone
    for tasks in all_tasks:
        results = pool.map(_function, tasks)
        for r1, r2 in zip(results, [_function(x) for x in tasks]):
            assert isclose(r1, r2)

        assert len(results) == len(tasks)

    # test map with callback
    for tasks in all_tasks:
        results = pool.map(_function, tasks, callback=_callback)
        for r1, r2 in zip(results, [_function(x) for x in tasks]):
            assert isclose(r1, r2)

        assert len(results) == len(tasks)

    pool.close()


if __name__ == '__main__':
    test_mpi_with_dill()



try:
    import numpy as np
    has_numpy = True
except ImportError:
    has_numpy = False

import pytest
from ..utils import batch_tasks


@pytest.mark.skipif(not has_numpy,
                    reason="Numpy is required to run this test")
def test_batch_tasks():
    tasks = batch_tasks(10, n_tasks=100)
    assert len(tasks) == 10
    assert tasks[0] == (0, 10)

    tasks = batch_tasks(10, n_tasks=100, args=(10,))
    assert len(tasks) == 10
    assert tasks[0] == ((0, 10), 10)

    tasks = batch_tasks(10, n_tasks=101, args=(99,))
    assert len(tasks) == 10

    # With array specified
    arr = np.random.random(size=123)
    tasks = batch_tasks(10, arr=arr, args=(99,))
    assert len(tasks) == 10

    arr = np.arange(100)
    tasks = batch_tasks(10, arr=arr, include_idx=False)
    assert np.all(tasks[0] == np.arange(10))

    tasks = batch_tasks(100, n_tasks=5)
    assert len(tasks) == 5

    with pytest.raises(ValueError):
        batch_tasks(0, n_tasks=100)

    with pytest.raises(ValueError):
        batch_tasks(100, n_tasks=0)

    with pytest.raises(ValueError):
        batch_tasks(100, n_tasks=100, arr=arr[:100])
